# A central place to define all environment variables used by Rasa

AUTH_TOKEN_ENV = "AUTH_TOKEN"
JWT_SECRET_ENV = "JWT_SECRET"
JWT_METHOD_ENV = "JWT_METHOD"
DEFAULT_JWT_METHOD = "HS256"
JWT_PRIVATE_KEY_ENV = "JWT_PRIVATE_KEY"

REMOTE_STORAGE_PATH_ENV = "REMOTE_STORAGE_PATH"
BUCKET_NAME_ENV = "BUCKET_NAME"
AWS_ENDPOINT_URL_ENV = "AWS_ENDPOINT_URL"
AZURE_CONTAINER_ENV = "AZURE_CONTAINER"
AZURE_ACCOUNT_NAME_ENV = "AZURE_ACCOUNT_NAME"
AZURE_ACCOUNT_KEY_ENV = "AZURE_ACCOUNT_KEY"
