import{p as e,f as o}from"./flowDb-956e92f1-ea63f85c.js";import{f as t,g as a}from"./styles-c10674c1-dc097813.js";import{u as i}from"./index-74e01d94.js";import"./graph-0a6f8466.js";import"./layout-e7fbc2bf.js";import"./index-3862675e-7610671a.js";import"./clone-8f9083bb.js";import"./edges-e0da2a9e-58676240.js";import"./createText-2e5e7dd3-088372e2.js";import"./line-a8aa457c.js";import"./array-9f3ba611.js";import"./path-53f90ab3.js";import"./channel-bf9cbb34.js";const M={parser:e,db:o,renderer:t,styles:a,init:r=>{r.flowchart||(r.flowchart={}),r.flowchart.arrowMarkerAbsolute=r.arrowMarkerAbsolute,i({flowchart:{arrowMarkerAbsolute:r.arrowMarkerAbsolute}}),t.setConf(r.flowchart),o.clear(),o.setGen("gen-2")}};export{M as diagram};
